"""Utilities for pytests for Reaxpro wrappers."""

from osp.models.catalytic.co_catalyticfoam import Boundary


def chem_bounds(value: "float") -> "Dict[Any, Any]":
    return [
        {"patch": "wedge1", "boundary_type": Boundary.wedge},
        {"patch": "wedge2", "boundary_type": Boundary.wedge},
        {"patch": "wedge3", "boundary_type": Boundary.wedge},
        {"patch": "wedge4", "boundary_type": Boundary.wedge},
        {"patch": "inlet", "boundary_type": Boundary.fixedValue, "value": value},
        {"patch": "outlet", "boundary_type": Boundary.zeroGradient},
        {"patch": "inertWall", "boundary_type": Boundary.zeroGradient},
        {
            "patch": "reactingWall",
            "boundary_type": Boundary.catalyticWall,
            "value": value,
        },
    ]


def u_bounds(value: "float") -> "Dict[Any, Any]":
    return [
        {"patch": "inlet", "boundary_type": Boundary.fixedValue, "value": value},
        {"patch": "outlet", "boundary_type": Boundary.zeroGradient},
        {"patch": "inertWall", "boundary_type": Boundary.fixedValue, "value": value},
        {"patch": "reactingWall", "boundary_type": Boundary.fixedValue, "value": value},
        {"patch": "wedge1", "boundary_type": Boundary.wedge},
        {"patch": "wedge2", "boundary_type": Boundary.wedge},
        {"patch": "wedge3", "boundary_type": Boundary.wedge},
        {"patch": "wedge4", "boundary_type": Boundary.wedge},
    ]


def t_bounds(value: "float") -> "Dict[Any, Any]":
    return [
        {"patch": "inlet", "boundary_type": Boundary.fixedValue, "value": value},
        {"patch": "outlet", "boundary_type": Boundary.zeroGradient},
        {"patch": "inertWall", "boundary_type": Boundary.zeroGradient},
        {
            "patch": "reactingWall",
            "boundary_type": Boundary.catalyticWall,
            "value": value,
        },
        {"patch": "wedge1", "boundary_type": Boundary.wedge},
        {"patch": "wedge2", "boundary_type": Boundary.wedge},
        {"patch": "wedge3", "boundary_type": Boundary.wedge},
        {"patch": "wedge4", "boundary_type": Boundary.wedge},
    ]


def p_bounds(value: "float") -> "Dict[Any, Any]":
    return [
        {"patch": "inlet", "boundary_type": Boundary.zeroGradient},
        {"patch": "outlet", "boundary_type": Boundary.fixedValue, "value": value},
        {"patch": "inertWall", "boundary_type": Boundary.zeroGradient},
        {"patch": "reactingWall", "boundary_type": Boundary.zeroGradient},
        {"patch": "wedge1", "boundary_type": Boundary.wedge},
        {"patch": "wedge2", "boundary_type": Boundary.wedge},
        {"patch": "wedge3", "boundary_type": Boundary.wedge},
        {"patch": "wedge4", "boundary_type": Boundary.wedge},
    ]



import logging
import os
import shutil
import subprocess  # nosec
import tempfile
from typing import TYPE_CHECKING

import osp.dictionaries.catalyticFoam as case

from .parsers import deep_replace, run_parser, serialize
from .settings import ReaxProSettings

if TYPE_CHECKING:
    from typing import Any, Dict, List

settings = ReaxProSettings()

CASES = os.path.dirname(case.__file__)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class CatalyticFoamEngine:
    def __init__(self, case: str, config: dict = {}):
        case_dir = os.path.join(CASES, case)
        if not config:
            runtime = tempfile.TemporaryDirectory().name
            tar = tempfile.NamedTemporaryFile().name
            config = dict(
                commands=[
                    dict(
                        exec="catalyticPimpleTurbulentFOAM",
                        args=None,
                        bashrc=str(settings.catalyticfoam_bash),
                    )
                ],
                directory=runtime,
                filename=tar,
            )
        self._parse_files = dict()
        self._config: dict = config
        self._current_process: subprocess.Popen = None
        logger.info("Engine instanciated with the following config: `%s`.", config)
        logger.info(
            "Will copy source %s to destination %s", case_dir, config["directory"]
        )
        shutil.copytree(case_dir, config["directory"])
        self._dir0 = os.path.join(config["directory"], "0")

    def add_to_parser(self, path: "str", jsonpath: "str", value: "Any") -> None:
        split = os.path.split(path)
        self._check_file(split)
        path = os.path.join(self._config["directory"], *split)
        logger.info(
            f"Will parse {value} into under query <{jsonpath}> into file {path}."
        )
        if path not in self._parse_files.keys():
            content = run_parser(path)
        else:
            content = self._parse_files[path]
        self._parse_files[path] = deep_replace(content, jsonpath, value)

    def run(self) -> None:
        self._check_consistency()
        for path, content in self._parse_files.items():
            with open(path, "w+") as file:
                file_content = serialize(content)
                file.write(file_content)
        for command in self._config["commands"]:
            cmd = self._prepare_command(command)
            logger.info(
                "Will run the following command under `%s`: `%s`.",
                self._config["directory"],
                cmd,
            )
            self._run(cmd)
        self._make_tarball()

    def _run(self, command: dict) -> None:
        with subprocess.Popen(
            f"bash -c '{command}'",  # nosec
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            cwd=self._config["directory"],
            shell=True,  # nosec
        ) as popen:
            for stdout_line in iter(popen.stdout.readline, ""):
                logger.info(stdout_line)
            for stderr_line in iter(popen.stderr.readline, ""):
                logger.error(stderr_line)
            self._current_process = popen
        return_code = popen.returncode
        if return_code:
            raise subprocess.CalledProcessError(return_code, command)

    def _prepare_command(self, command: dict) -> str:
        cmd = [command["exec"], command.get("args") or str()]
        if command.get("bashrc"):
            cmd = [".", command.get("bashrc"), "&&"] + cmd
        return " ".join(cmd)

    def _make_tarball(self) -> None:
        logger.info(
            "Make tarball in file `%s` of directory `%s`.",
            self.tarball,
            self._config["directory"],
        )
        shutil.make_archive(self._config["filename"], "tar", self._config["directory"])

    def _check_file(self, filename: "List[str]") -> None:
        filepath = os.path.join(self._config["directory"], *filename)
        templatepath = os.path.join(CASES, "templates", filename[-1])
        if not os.path.exists(filepath):
            logger.info(
                f"File for {filepath} does not exist yet. Will create from template"
            )
            if not os.path.exists(templatepath):
                templatepath = os.path.join(CASES, "templates", "template_chem")
                logger.info(
                    f"Specific template for {filename[-1]} does not exist. Will use {templatepath}."
                )
            shutil.copy(templatepath, filepath)

    def _check_consistency(self) -> None:
        excluded = [
            exclude.strip()
            for exclude in settings.catalyticfoam_excluded_check_files.split("|")
        ]
        logger.info(
            f"Detected to exclude the following file while checking boundary conditions: {excluded}"
        )
        for file in os.listdir(self._dir0):
            path = os.path.join(self._dir0, file)
            path0 = os.path.join("0", file)
            if (
                os.path.isfile(path)
                and path0 not in excluded
                and path not in self._parse_files.keys()
            ):
                try:
                    self._parse_files[path] = run_parser(path)
                except Exception:
                    logger.info(
                        f"{path0} does not seem to be a file which can be parsed. Will skip this file."
                    )
            else:
                logger.info(
                    f"{path0} was already parsed or is exlcuded. Will skip this file."
                )
        boundaries = self._get_defined_boundaries()
        self._check_defaults(boundaries)
        self._check_bounds(boundaries)

    def _get_defined_boundaries(self):
        boundaries = run_parser(
            os.path.join(self._config["directory"], "constant", "polyMesh", "boundary")
        )
        boundaries_def = [
            bounds
            for key, value in boundaries.items()
            if key.isnumeric()
            for bounds in value
        ]
        return {key: value for bound in boundaries_def for key, value in bound.items()}

    def _check_defaults(self, boundaries: "Dict[Any,Any]"):
        for rawpath in settings.catalyticfoam_default_files.split("|"):
            path = os.path.join(self._config["directory"], rawpath.strip())
            content = self._parse_files.get(path)
            if not content:
                raise ValueError(
                    f"{path} was not found in the buffer. Please check `excluded_check_files` in the settings."
                )
            boundaryField = content.get("boundaryField")
            if not boundaryField:
                raise ValueError(f"boundaryField not found in {path}")
            for boundary in boundaries:
                if boundary != "defaultFaces" and boundary not in boundaryField.keys():
                    default = self._find_matches(boundary)
                    if not default:
                        raise ValueError(
                            f"Needed boundary '{boundary}', was not defined in `default_bounds`. Please check the settings."
                        )
                    logger.info(
                        f"""Will place `{default}` under needed boundary `{boundary}` in file {path}`,
                        since it was not defined there yet."""
                    )
                    boundaryField[boundary] = default

    def _check_bounds(self, boundaries: "Dict[Any,Any]"):
        for path, content in self._parse_files.items():
            if "0" in path.split(os.path.sep):
                boundaryField = content.get("boundaryField")
                if not boundaryField:
                    raise ValueError(f"boundaryField not found in {path}")
                for boundary in boundaries:
                    if (
                        boundary != "defaultFaces"
                        and boundary not in boundaryField.keys()
                    ):
                        raise ValueError(f"Boundary '{boundary}', must be in {path}.")
                for key, values in boundaryField.items():
                    if "type" not in values.keys():
                        raise ValueError(
                            f"Type of boundary must be specified through 'key' in {path}/{key}/type."
                        )
                logger.info(
                    f"File {path} was validated. All boundaries are set correctly."
                )

    def _find_matches(self, key):
        for match, content in settings.catalyticfoam_default_bounds.items():
            if key.startswith(match):
                return content

    @property
    def exit_code(cls):
        return cls._current_process.returncode

    @property
    def tarball(cls):
        return f"{cls._config['filename']}.tar"



import os
import shutil
import tempfile

from osp.core.namespaces import cuba, emmo
from osp.models.catalytic.utils import make_arcp
from osp.wrappers.simcatalyticfoam.parsers import run_parser
from osp.wrappers.simcatalyticfoam.simcatalyticfoam import SimCatalyticFoamSession
from osp.wrappers.simcatalyticfoam.utils import (
    _make_boundary_field,
    _make_internal_field,
)

p_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "binary",
        "class": "volScalarField",
        "location": '"0"',
        "object": "p",
    },
    "dimensions": [1, -1, -2, 0, 0, 0, 0],
    "internalField": ["uniform", 100000],
    "boundaryField": {
        "inlet": {"type": "zeroGradient"},
        "outlet": {"type": "fixedValue", "value": ["uniform", 100000]},
        "inertWall": {"type": "zeroGradient"},
        "reactingWall": {"type": "zeroGradient"},
        "wedge1": {"type": "wedge"},
        "wedge2": {"type": "wedge"},
        "wedge3": {"type": "wedge"},
        "wedge4": {"type": "wedge"},
    },
}

CO_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "binary",
        "class": "volScalarField",
        "location": '"0"',
        "object": "CO",
    },
    "dimensions": [0, 0, 0, 0, 0, 0, 0],
    "internalField": ["uniform", 9.45945945945946e-06],
    "boundaryField": {
        "inlet": {"type": "fixedValue", "value": ["uniform", 9.45945945945946e-06]},
        "outlet": {"type": "zeroGradient"},
        "inertWall": {
            "type": "zeroGradient",
        },
        "reactingWall": {
            "type": "catalyticWall",
            "value": ["uniform", 9.45945945945946e-06],
        },
        "wedge1": {"type": "wedge"},
        "wedge2": {"type": "wedge"},
        "wedge3": {"type": "wedge"},
        "wedge4": {"type": "wedge"},
    },
}


n2_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "binary",
        "class": "volScalarField",
        "location": '"0"',
        "object": "N2",
    },
    "dimensions": [0, 0, 0, 0, 0, 0, 0],
    "internalField": ["uniform", 0.567558108108108],
    "boundaryField": {
        "inlet": {"type": "fixedValue", "value": ["uniform", 0.567558108108108]},
        "outlet": {"type": "zeroGradient"},
        "inertWall": {
            "type": "zeroGradient",
        },
        "reactingWall": {
            "type": "catalyticWall",
            "value": ["uniform", 0.567558108108108],
        },
        "wedge1": {"type": "wedge"},
        "wedge2": {"type": "wedge"},
        "wedge3": {"type": "wedge"},
        "wedge4": {"type": "wedge"},
    },
}


O2_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "binary",
        "class": "volScalarField",
        "location": '"0"',
        "object": "O2",
    },
    "dimensions": [0, 0, 0, 0, 0, 0, 0],
    "internalField": ["uniform", 0.432432432432432],
    "boundaryField": {
        "inlet": {"type": "fixedValue", "value": ["uniform", 0.432432432432432]},
        "outlet": {"type": "zeroGradient"},
        "inertWall": {
            "type": "zeroGradient",
        },
        "reactingWall": {
            "type": "catalyticWall",
            "value": ["uniform", 0.4324324324324328],
        },
        "wedge1": {"type": "wedge"},
        "wedge2": {"type": "wedge"},
        "wedge3": {"type": "wedge"},
        "wedge4": {"type": "wedge"},
    },
}

U_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "binary",
        "class": "volVectorField",
        "location": '"0"',
        "object": "U",
    },
    "boundaryField": {
        "inertWall": {"type": "fixedValue", "value uniform": [0, 0, 1]},
        "inlet": {"type": "fixedValue", "value uniform": [0, 0, 1]},
        "outlet": {"type": "zeroGradient"},
        "reactingWall": {"type": "fixedValue", "value uniform": [0, 0, 1]},
        "wedge1": {"type": "wedge"},
        "wedge2": {"type": "wedge"},
        "wedge3": {"type": "wedge"},
        "wedge4": {"type": "wedge"},
    },
    "dimensions": [0, 1, -1, 0, 0, 0, 0],
    "internalField": ["uniform", [0, 0, 1]],
}

T_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "binary",
        "class": "volScalarField",
        "location": '"0"',
        "object": "T",
    },
    "dimensions": [0, 0, 0, 1, 0, 0, 0],
    "internalField": ["uniform", 900],
    "boundaryField": {
        "inlet": {"type": "fixedValue", "value": ["uniform", 900]},
        "outlet": {"type": "zeroGradient"},
        "inertWall": {"type": "zeroGradient"},
        "reactingWall": {"type": "catalyticWall", "value": ["uniform", 900]},
        "wedge1": {"type": "wedge"},
        "wedge2": {"type": "wedge"},
        "wedge3": {"type": "wedge"},
        "wedge4": {"type": "wedge"},
    },
}


theta_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "binary",
        "class": "volScalarField",
        "location": '"0"',
        "object": "thetaDefault",
    },
    "dimensions": [0, 0, 0, 0, 0, 0, 0],
    "internalField": ["uniform", 0],
    "boundaryField": {
        "inlet": {"type": "zeroGradient"},
        "outlet": {"type": "zeroGradient"},
        "inertWall": {"type": "zeroGradient"},
        "reactingWall": {"type": "fixedValue", "value": ["uniform", 0]},
        "wedge1": {"type": "wedge"},
        "wedge2": {"type": "wedge"},
        "wedge3": {"type": "wedge"},
        "wedge4": {"type": "wedge"},
    },
}

y_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "ascii",
        "class": "volScalarField",
        "location": '"1"',
        "object": "Ydefault",
    },
    "dimensions": [0, 0, 0, 0, 0, 0, 0],
    "internalField": ["uniform", 0],
    "boundaryField": {
        "reactingWall": {"type": "catalyticWall", "value": ["uniform", 0]},
        "inertWall": {"type": "zeroGradient"},
        "inlet": {"type": "fixedValue", "value": ["uniform", 0]},
        "outlet": {"type": "zeroGradient"},
        "wedge1": {"type": "wedge"},
        "wedge2": {"type": "wedge"},
        "wedge3": {"type": "wedge"},
        "wedge4": {"type": "wedge"},
    },
}

turbulence_expected = {
    "FoamFile": {
        "version": 2.0,
        "format": "ascii",
        "class": "dictionary",
        "location": '"constant"',
        "object": "turbulenceProperties",
    },
    "simulationType": "laminar",
}

solver_expected = {
    "FoamFile": {
        "version": 2.3,
        "format": "binary",
        "class": "dictionary",
        "location": '"constant"',
        "object": "solverOptions",
    },
    "MachineLearning": {
        "MLSurrogate": "on",
        "MachineLearningFile": '"ml_ExtraTrees_forCFD.pkl"',
    },
    "PhysicalModel": {
        "strangAlgorithm": '"TransportReactionMomentum"',
        "homogeneousReactions": "off",
        "heterogeneousReactions": "on",
        "alfaCatalyst": 1.5e-05,
        "Sct": 0.7,
        "Prt": 0.7,
        "catalyticWalls": ["reactingWall"],
        "energyEquation": "on",
        "reactionHeatFromHeterogeneousReactions": "on",
        "constPressureBatchReactor": True,
        "massDiffusionInEnergyEquation": "on",
        "diffusivityModel": "fick-multi-component",
        "includeDpDt": "off",
        "soretEffect": "off",
    },
    "OdeHomogeneous": {
        "odeSolver": '"OpenSMOKE"',
        "relTolerance": 1e-07,
        "absTolerance": 1e-12,
        "maximumOrder": 5,
        "fullPivoting": False,
    },
    "OdeHeterogeneous": {
        "odeSolver": '"OpenSMOKE"',
        "relTolerance": 1e-07,
        "absTolerance": 1e-12,
        "maximumOrder": 5,
        "fullPivoting": False,
    },
    "LewisNumbers": {"O2": 1.0, "H2": 1.1, "H2O": 1.2, "N2": 1.3},
    "#include": '"isatOptions"',
}


def test_foam():
    config = dict(
        commands=[
            dict(
                exec="echo",
                args="hello world",
            )
        ],
        directory=tempfile.TemporaryDirectory().name,
        filename=f"{tempfile.NamedTemporaryFile().name}.tar",
    )

    #############################################################################################

    #                                   CO

    ############################################################################################

    CO_composition = emmo.ChemicalComposition()

    CO = emmo.ChemicalSpecies()

    iri = make_arcp("0/CO", query=dict(jsonpath=[["FoamFile", "object"]]))
    CO_symbol = emmo.Symbol(iri=iri, hasSymbolData="CO")

    CO.add(CO_symbol)  # Check alternative relation

    CO_composition.add(CO, rel=emmo.hasSpatialDirectPart)

    val = 9.45945945945946e-06

    ###### internalField

    mass_fraction = _make_internal_field("0/CO", val, emmo.MassFraction)

    CO_composition.add(mass_fraction, rel=emmo.hasSpatialDirectPart)

    ###### BoundaryField

    ### reactingwall
    bc = _make_boundary_field(
        "0/CO",
        "reactingWall",
        emmo.CatalyticWall,
        quantity=emmo.MassFraction,
        value=val,
    )
    CO_composition.add(bc, rel=emmo.hasBoundaryField)

    ### inlet
    bc = _make_boundary_field(
        "0/CO", "inlet", emmo.FixedValue, quantity=emmo.MassFraction, value=val
    )
    CO_composition.add(bc, rel=emmo.hasBoundaryField)

    ### inertWall, outlet, wedges
    for patch, condition in [
        ("inertWall", emmo.ZeroGradient),
        ("outlet", emmo.ZeroGradient),
        ("wedge1", emmo.Wedge),
        ("wedge2", emmo.Wedge),
        ("wedge3", emmo.Wedge),
        ("wedge4", emmo.Wedge),
    ]:
        bc = _make_boundary_field("0/CO", patch, condition)
        CO_composition.add(bc, rel=emmo.hasBoundaryField)

    ######################################################################################

    #                                   Nitrogen

    ######################################################################################

    n2_composition = emmo.ChemicalComposition()

    n2 = emmo.ChemicalSpecies()

    iri = make_arcp("0/N2", query=dict(jsonpath=[["FoamFile", "object"]]))
    n2_symbol = emmo.Symbol(iri=iri, hasSymbolData="N2")

    n2.add(n2_symbol)  # Check alternative relation

    n2_composition.add(n2, rel=emmo.hasSpatialDirectPart)

    val = 0.567558108108108

    ###### internalField

    mass_fraction = _make_internal_field("0/N2", val, emmo.MassFraction)
    n2_composition.add(mass_fraction, rel=emmo.hasSpatialDirectPart)

    ###### BoundaryField

    ### reactingwall
    bc = _make_boundary_field(
        "0/N2",
        "reactingWall",
        emmo.CatalyticWall,
        value=val,
        quantity=emmo.MassFraction,
    )
    n2_composition.add(bc, rel=emmo.hasBoundaryField)

    ### inlet
    bc = _make_boundary_field(
        "0/N2", "inlet", emmo.FixedValue, value=val, quantity=emmo.MassFraction
    )
    n2_composition.add(bc, rel=emmo.hasBoundaryField)

    ### inertWall, outlet, wedges
    for patch, condition in [
        ("inertWall", emmo.ZeroGradient),
        ("outlet", emmo.ZeroGradient),
        ("wedge1", emmo.Wedge),
        ("wedge2", emmo.Wedge),
        ("wedge3", emmo.Wedge),
        ("wedge4", emmo.Wedge),
    ]:
        bc = _make_boundary_field("0/N2", patch, condition)
        n2_composition.add(bc, rel=emmo.hasBoundaryField)

    ######################################################################################

    #                                   Oxygen

    ######################################################################################

    O2_composition = emmo.ChemicalComposition()

    O2 = emmo.ChemicalSpecies()

    iri = make_arcp("0/O2", query=dict(jsonpath=[["FoamFile", "object"]]))
    O2_symbol = emmo.Symbol(iri=iri, hasSymbolData="O2")

    O2.add(O2_symbol)  # Check alternative relation

    O2_composition.add(CO, rel=emmo.hasSpatialDirectPart)

    ###### internalField

    val = 0.432432432432432

    mass_fraction = _make_internal_field("0/O2", val, emmo.MassFraction)

    O2_composition.add(mass_fraction, rel=emmo.hasSpatialDirectPart)

    ###### BoundaryField

    ### reactingwall
    bc = _make_boundary_field(
        "0/O2",
        "reactingWall",
        emmo.CatalyticWall,
        value=val,
        quantity=emmo.MassFraction,
    )
    O2_composition.add(bc, rel=emmo.hasBoundaryField)

    ### inlet
    bc = _make_boundary_field(
        "0/O2", "inlet", emmo.FixedValue, value=val, quantity=emmo.MassFraction
    )
    O2_composition.add(bc, rel=emmo.hasBoundaryField)

    # inertwall, outlet, wedges
    for patch, condition in [
        ("inertWall", emmo.ZeroGradient),
        ("outlet", emmo.ZeroGradient),
        ("wedge1", emmo.Wedge),
        ("wedge2", emmo.Wedge),
        ("wedge3", emmo.Wedge),
        ("wedge4", emmo.Wedge),
    ]:
        bc = _make_boundary_field("0/O2", patch, condition)
        O2_composition.add(bc, rel=emmo.hasBoundaryField)

    ######################################################################################

    #                                   Pressure

    ######################################################################################

    val = 100000

    ##### internalField

    pressure = _make_internal_field("0/p", val, emmo.Pressure)

    ##### BoundaryField

    ### outlet
    bc = _make_boundary_field(
        "0/p", "outlet", emmo.FixedValue, value=val, quantity=emmo.Pressure
    )
    pressure.add(bc, rel=emmo.hasBoundaryField)

    ### reactingWall, inertwall, inlet, wedges
    for patch, condition in [
        ("inertWall", emmo.ZeroGradient),
        ("reactingWall", emmo.ZeroGradient),
        ("inlet", emmo.ZeroGradient),
        ("wedge1", emmo.Wedge),
        ("wedge2", emmo.Wedge),
        ("wedge3", emmo.Wedge),
        ("wedge4", emmo.Wedge),
    ]:
        bc = _make_boundary_field("0/p", patch, condition)
        pressure.add(bc, rel=emmo.hasBoundaryField)

    ######################################################################################

    #                                   Temperature

    ######################################################################################

    val = 900

    ###### internalField

    temperature = _make_internal_field("0/T", val, emmo.ThermodynamicTemperature)

    ###### boundaryField

    ### reactingwall
    bc = _make_boundary_field(
        "0/T",
        "reactingWall",
        emmo.CatalyticWall,
        value=val,
        quantity=emmo.ThermodynamicTemperature,
    )
    temperature.add(bc, rel=emmo.hasBoundaryField)

    ### inlet
    bc = _make_boundary_field(
        "0/T",
        "inlet",
        emmo.FixedValue,
        value=val,
        quantity=emmo.ThermodynamicTemperature,
    )
    temperature.add(bc, rel=emmo.hasBoundaryField)

    ### inertwall, outlet, wedges
    for patch, condition in [
        ("inertWall", emmo.ZeroGradient),
        ("outlet", emmo.ZeroGradient),
        ("wedge1", emmo.Wedge),
        ("wedge2", emmo.Wedge),
        ("wedge3", emmo.Wedge),
        ("wedge4", emmo.Wedge),
    ]:
        bc = _make_boundary_field("0/T", patch, condition)
        temperature.add(bc, rel=emmo.hasBoundaryField)

    ######################################################################################

    #                                   Velocity

    ######################################################################################

    val = [0, 0, 1]

    ###### internalField

    velocity = _make_internal_field("0/U", val, emmo.Velocity)

    ###### boundaryField

    # reactingWall, inlet, inertWall, wedges
    for patch in ["reactingWall", "inlet", "inertWall"]:
        bc = _make_boundary_field(
            "0/U", patch, emmo.FixedValue, value=val, quantity=emmo.Velocity
        )
        velocity.add(bc, rel=emmo.hasBoundaryField)

    # outlet, wedges
    for patch, condition in [
        ("outlet", emmo.ZeroGradient),
        ("wedge1", emmo.Wedge),
        ("wedge2", emmo.Wedge),
        ("wedge3", emmo.Wedge),
        ("wedge4", emmo.Wedge),
    ]:
        bc = _make_boundary_field("0/U", patch, condition)
        velocity.add(bc, rel=emmo.hasBoundaryField)

    ######################################################################################

    # Create calculation

    calc = emmo.ContinuumCalculation()

    # Solver options

    iri = make_arcp(
        "constant/solverOptions",
        query=dict(jsonpath=[["PhysicalModel", "alfaCatalyst"]]),
    )
    catalyst_amount = emmo.CatalystAmount()
    catalyst_amount_value = emmo.Real(iri=iri, hasNumericalData=1.5e-5)
    catalyst_amount.add(catalyst_amount_value, rel=emmo.hasQuantityValue)

    iri = make_arcp(
        "constant/solverOptions",
        query=dict(jsonpath=[["PhysicalModel", "energyEquation"]]),
    )
    energy_eq = emmo.EnergyEquation(iri=iri)

    iri = make_arcp(
        "constant/solverOptions",
        query=dict(jsonpath=[["PhysicalModel", "diffusivityModel"]]),
    )
    diffusivity_model = emmo.FicksDiffusion(iri=iri)

    iri = make_arcp(
        "constant/turbulenceProperties", query=dict(jsonpath=[["simulationType"]])
    )
    turbulent_model = emmo.LaminarModel(iri=iri)

    model = emmo.ContinuumModel()
    model.add(
        energy_eq,
        diffusivity_model,
        turbulent_model,
        rel=emmo.hasSpatialDirectPart,
    )

    calc.add(
        CO_composition,
        n2_composition,
        O2_composition,
        pressure,
        temperature,
        velocity,
        model,
        catalyst_amount,
        rel=emmo.hasCalculationInput,
    )

    session = SimCatalyticFoamSession(config=config)
    wrapper = cuba.Wrapper(session=session)
    wrapper.add(calc, rel=cuba.relationship)
    session.run()

    p = run_parser(os.path.join(session.config["directory"], "0", "p"))
    T = run_parser(os.path.join(session.config["directory"], "0", "T"))
    U = run_parser(os.path.join(session.config["directory"], "0", "U"))
    CO = run_parser(os.path.join(session.config["directory"], "0", "CO"))
    n2 = run_parser(os.path.join(session.config["directory"], "0", "N2"))
    y = run_parser(os.path.join(session.config["directory"], "0", "Ydefault"))
    theta = run_parser(os.path.join(session.config["directory"], "0", "thetaDefault"))
    turbulence = run_parser(
        os.path.join(session.config["directory"], "constant", "turbulenceProperties")
    )
    solver = run_parser(
        os.path.join(session.config["directory"], "constant", "solverOptions")
    )

    assert p == p_expected

    assert U == U_expected

    assert T == T_expected

    assert CO == CO_expected

    assert n2 == n2_expected

    assert theta == theta_expected

    assert y == y_expected

    assert solver == solver_expected

    assert turbulence == turbulence_expected

    assert session.exit_code == 0



"""Full multiscale workflow from atomistic to continuum."""

import tempfile
import warnings

from pydantic import Field
from pydantic.dataclasses import dataclass

from osp.core.namespaces import crystallography, emmo
from osp.core.session import CoreSession
from osp.core.utils import export_cuds
from osp.models.catalytic.co_catalyticfoam import COCatalyticFOAMModel
from osp.models.multiscale.co_pt111_meso import (
    BindingSite,
    COMolarFractionRange,
    PESExploration,
    ZGBModel,
)
from osp.models.utils.general import _get_example_json, get_upload
from osp.models.zacros.co_pyzacros import COpyZacrosModel

STANDARD_XYZ = [
    ("4442d5c3-4b61-4b13-9bbb-fdf942776ca6", "CO_ads+Pt111.xyz"),
]


@dataclass
class ContinuumModel(COCatalyticFOAMModel):
    """Continuum model for full multiscale workflow."""

    class Config:
        """Pydantic model."""

        exclude = {"species_from_upload"}


@dataclass
class ZacrosModel(COpyZacrosModel):
    """Continuum model for full multiscale workflow."""

    class Config:
        """Pydantic model."""

        exclude = {"apd"}


@dataclass
class COPt111FullscaleModel:
    """Pydantic model describing the Full scale for CO oxidation."""

    pes_exploration: PESExploration = Field(
        ..., description="AMS data model for PESExploration."
    )
    binding_site: BindingSite = Field(
        ...,
        description="""data model for binding site calculation
        based on the previous PESExploraion.""",
    )
    zgb_model: ZGBModel = Field(..., description="ZGB model for mesoscopic scale.")

    adp: COMolarFractionRange = Field(
        ...,
        description="""Molar fractions of CO
        for the adaptive design procedure""",
    )

    catalyticfoam: ContinuumModel = Field(
        ..., description="Data model for continuum scale."
    )

    def __post_init_post_parse__(self):
        with CoreSession() as session:
            workflow = emmo.Workflow()
            apd = emmo.AdaptiveDesignProcedure()
            apd.add(self.adp.cuds, rel=emmo.hasInput)
            for oclass in [
                emmo.ForceFieldIdentifierString,
                emmo.Solver,
                emmo.FixedRegion,
                emmo.MaximumEnergy,
                emmo.NeighborCutoff,
                emmo.ReferenceRegion,
                emmo.RandomSeed,
                emmo.MolecularGeometry,
                crystallography.UnitCell,
            ]:
                input_cuds = self.pes_exploration.cuds.get(
                    oclass=oclass, rel=emmo.hasInput
                )
                self.binding_site.cuds.add(input_cuds.pop(), rel=emmo.hasInput)
            self.pes_exploration.cuds.add(
                self.binding_site.cuds, rel=emmo.hasSpatialNext
            )
            self.binding_site.cuds.add(self.zgb_model.cuds, rel=emmo.hasSpatialNext)
            self.zgb_model.cuds.add(apd, rel=emmo.hasSpatialNext)
            apd.add(self.catalyticfoam.cuds, rel=emmo.hasSpatialNext)
            workflow.add(self.pes_exploration.cuds, rel=emmo.hasSpatialFirst)
            workflow.add(self.binding_site.cuds, rel=emmo.hasSpatialDirectPart)
            workflow.add(self.zgb_model.cuds, rel=emmo.hasSpatialDirectPart)
            workflow.add(apd, rel=emmo.hasSpatialDirectPart)
            workflow.add(self.catalyticfoam.cuds, rel=emmo.hasSpatialLast)

        file = tempfile.NamedTemporaryFile(suffix=".ttl", delete=False)
        export_cuds(session, file.name)
        self._file = file.name
        try:
            self._uuid = get_upload(file)
        except Exception as error:
            self._uuid = None
            message = (
                message
            ) = f"The graph of the model could not be stored at the minio-instance: {error.args}"
            warnings.warn(message)
        self._session = session

    @property
    def session(self) -> "CoreSession":
        return self._session

    @property
    def cuds(cls):
        return cls._session.load(cls._session.root).first()

    @property
    def uuid(cls):
        return cls._uuid

    @property
    def file(cls):
        return cls._file

    class Config:
        """Pydantic Config"""

        schema_extra = {
            "example": _get_example_json("co_pt111_full.json", STANDARD_XYZ)
        }


@dataclass
class COPt111FromMesoScaleModel:
    """Pydantic model describing data model for
    CO oxidation beginning on the mescoscopic scale."""

    zgb_model: ZacrosModel = Field(..., description="ZGB model for mesoscopic scale.")

    adp: COMolarFractionRange = Field(
        ...,
        description="""Molar fractions of CO
        for the adaptive design procedure""",
    )

    catalyticfoam: ContinuumModel = Field(
        ..., description="Data model for continuum scale."
    )

    def __post_init_post_parse__(self):
        with CoreSession() as session:
            workflow = emmo.Workflow()
            self.zgb_model.cuds.add(apd, rel=emmo.hasSpatialNext)
            apd.add(self.catalyticfoam.cuds, rel=emmo.hasSpatialNext)
            apd = emmo.AdaptiveDesignProcedure()
            apd.add(self.adp.cuds, rel=emmo.hasInput)
            workflow.add(self.zgb_model.cuds, rel=emmo.hasSpatialFirst)
            workflow.add(apd, rel=emmo.hasSpatialDirectPart)
            workflow.add(self.catalyticfoam.cuds, rel=emmo.hasSpatialLast)

        file = tempfile.NamedTemporaryFile(suffix=".ttl", delete=False)
        export_cuds(session, file.name)
        self._file = file.name
        try:
            self._uuid = get_upload(file)
        except Exception as error:
            self._uuid = None
            message = (
                message
            ) = f"The graph of the model could not be stored at the minio-instance: {error.args}"
            warnings.warn(message)
        self._session = session

    @property
    def session(self) -> "CoreSession":
        return self._session

    @property
    def cuds(cls):
        return cls._session.load(cls._session.root).first()

    @property
    def uuid(cls):
        return cls._uuid

    @property
    def file(cls):
        return cls._file

    class Config:
        """Pydantic Config"""

        schema_extra = {
            "example": _get_example_json("co_pt111_from_meso.json", STANDARD_XYZ)
        }



"""Model settings for OSP Pydantic Models"""
from pydantic import BaseSettings, Field, SecretStr


class ModelSettings(BaseSettings):
    """Main configuration for based on redis-settings"""

    minio_user: SecretStr = Field("rootname", description="User to MinIO instance.")
    minio_password: SecretStr = Field(
        "foobar123", description="Password to MinIO instance."
    )
    minio_endpoint: str = Field(
        "localhost:9000",
        description="Resolvable endpoint to contact MinIO instance.",
    )
    download_timeout: int = Field(
        10, description="Timeout when a file needs to be downloaded via http."
    )

    class Config:
        """Pydantic Config for ReaxProSettings."""

        env_prefix = "REAXPRO_"



def test_read_case():
    from osp.models.catalytic.utils import _read_patches

    expected = [
        "wedge1",
        "wedge2",
        "wedge3",
        "wedge4",
        "reactingWall",
        "inlet",
        "outlet",
        "inertWall",
    ]

    assert sorted(expected) == sorted(_read_patches())



import json
import os
from pathlib import Path
from typing import Any, Dict, List
from urllib.parse import quote, urlencode

import joblib
import requests
from arcp import arcp_random, is_arcp_uri
from arcp.parse import ARCPParseResult, parse_arcp

import osp.dictionaries as cases
from osp.core.cuds import Cuds
from osp.models.utils.general import get_download, get_upload
from osp.wrappers.simcatalyticfoam.parsers import run_parser
from osp.wrappers.simcatalyticfoam.settings import ReaxProSettings

CASES = os.path.dirname(cases.__file__)
settings = ReaxProSettings()


def make_arcp(*args, **kwargs):
    if kwargs.get("query"):
        query = kwargs.pop("query")
        for key, value in query.items():
            query[key] = [quote(".".join(item)) for item in value]
        query = urlencode(query, doseq=True)
    return arcp_random(*args, **kwargs, query=query)


def check_arcp(cuds: Cuds) -> bool:
    return is_arcp_uri(cuds.iri)


def wrap_arcp(cuds: Cuds) -> ARCPParseResult:
    return parse_arcp(cuds.iri)


def _import_from_pkl(
    file_path: Path = settings.catalyticfoam_standard_pkl,
) -> List[str]:
    _, _, _, tabVarParam = joblib.load(file_path)
    return tabVarParam["names"]


def _read_from_case(
    rel_file_path: Path, case_path: Path = settings.catalyticfoam_standard_case
) -> Dict[Any, Any]:
    file_path = os.path.join(case_path, rel_file_path)
    return run_parser(file_path)


def _read_patches(case_path: Path = settings.catalyticfoam_standard_case) -> List[str]:
    path = os.path.join("constant", "polyMesh", "boundary")
    boundaries = _read_from_case(path, case_path=case_path)
    boundaries_def = [
        bounds
        for key, value in boundaries.items()
        if key.isnumeric()
        for bounds in value
    ]
    return [
        key for bound in boundaries_def for key in bound.keys() if key != "defaultFaces"
    ]



import os
from io import BytesIO
from typing import List
from uuid import uuid4

import pytest

PATH = os.path.dirname(__file__)
TEST_PATH = os.path.join(PATH, "test_files")


def test_co():
    from osp.models.catalytic.co_catalyticfoam import COCatalyticFOAMModel

    from .utils import chem_bounds, p_bounds, t_bounds, u_bounds

    config = {
        "chemical_species": [
            {"composition": "CO", "mass_fraction": 1, "boundaries": chem_bounds(1)}
        ],
        "velocity": {
            "value": [0.1, 0.1, 0.1],
            "boundaries": u_bounds([0.1, 0.1, 0.1]),
        },
        "temperature": {"value": 300.0, "boundaries": t_bounds(300.0)},
        "pressure": {"value": 1000.0, "boundaries": p_bounds(1000.0)},
        "catalyst_amount": 1.5e-5,
    }
    model = COCatalyticFOAMModel(**config)
    assert model.chemical_species[0].composition == "CO"
    assert model.velocity.value[0] == 0.1
    assert model.velocity.value[1] == 0.1
    assert model.velocity.value[2] == 0.1
    assert model.velocity.value[2] == 0.1


def test_wrong_velocity_vec():
    from pydantic.error_wrappers import ValidationError

    from osp.models.catalytic.co_catalyticfoam import COCatalyticFOAMModel

    from .utils import chem_bounds, p_bounds, t_bounds, u_bounds

    config_error = {
        "chemical_species": [
            {"composition": "CO", "mass_fraction": 1, "boundaries": chem_bounds(1)},
        ],
        "velocity": {
            "value": [0.1, 0.1, 0.1, 0.1, 0.1],
            "boundaries": u_bounds([0.1, 0.1, 0.1]),
        },
        "temperature": {"value": 300.0, "boundaries": t_bounds(300.0)},
        "pressure": {"value": 1000.0, "boundaries": p_bounds(1000.0)},
        "catalyst_amount": 1.5e-5,
    }
    with pytest.raises(
        ValidationError, match="ensure this value has at most 3 items"
    ) as err:
        COCatalyticFOAMModel(**config_error)


def test_velocity_vec_too_short():
    from pydantic.error_wrappers import ValidationError

    from osp.models.catalytic.co_catalyticfoam import COCatalyticFOAMModel

    from .utils import chem_bounds, p_bounds, t_bounds, u_bounds

    config_error = {
        "chemical_species": [
            {"composition": "CO", "mass_fraction": 1, "boundaries": chem_bounds(1)},
        ],
        "velocity": {
            "value": [0.1, 0.1],
            "boundaries": u_bounds([0.1, 0.1, 0.1]),
        },
        "temperature": {"value": 300.0, "boundaries": t_bounds(300.0)},
        "pressure": {"value": 1000.0, "boundaries": p_bounds(1000.0)},
        "catalyst_amount": 1.5e-5,
    }
    with pytest.raises(
        ValidationError, match="ensure this value has at least 3 items"
    ) as err:
        COCatalyticFOAMModel(**config_error)


def test_wrong_composition():
    from osp.models.catalytic.co_catalyticfoam import COCatalyticFOAMModel

    from .utils import chem_bounds, p_bounds, t_bounds, u_bounds

    config_error = {
        "chemical_species": [
            {"composition": "COA", "mass_fraction": 1, "boundaries": chem_bounds(1)},
        ],
        "velocity": {
            "value": [0.1, 0.1, 0.1],
            "boundaries": u_bounds([0.1, 0.1, 0.1]),
        },
        "temperature": {"value": 300.0, "boundaries": t_bounds(300.0)},
        "pressure": {"value": 1000.0, "boundaries": p_bounds(1000.0)},
        "catalyst_amount": 1.5e-5,
    }
    with pytest.raises(ValueError, match="Invalid composition") as err:
        COCatalyticFOAMModel(**config_error)


def test_co_from_pkl():
    from pydantic.error_wrappers import ValidationError

    from osp.models.catalytic.co_catalyticfoam import COCatalyticFOAMModel
    from osp.wrappers.simcatalyticfoam.settings import ReaxProSettings

    settings = ReaxProSettings()

    from .conftest import store_file
    from .utils import chem_bounds, p_bounds, t_bounds, u_bounds

    pkl = os.path.join(TEST_PATH, "test.pkl")
    with open(pkl, "rb") as file:
        key = str(store_file(file))

    config = {
        "chemical_species": [
            {"composition": "CO", "mass_fraction": 1, "boundaries": chem_bounds(1)}
        ],
        "species_from_upload": key,
        "velocity": {
            "value": [0.1, 0.1, 0.1],
            "boundaries": u_bounds([0.1, 0.1, 0.1]),
        },
        "temperature": {"value": 300.0, "boundaries": t_bounds(300.0)},
        "pressure": {"value": 1000.0, "boundaries": p_bounds(1000.0)},
        "catalyst_amount": 1.5e-5,
    }
    model = COCatalyticFOAMModel(**config)
    assert model.chemical_species[0].composition == "CO"

    config_error = {
        "chemical_species": [
            {"composition": "COA", "mass_fraction": 1, "boundaries": chem_bounds(1)}
        ],
        "species_from_upload": key,
        "velocity": {
            "value": [0.1, 0.1, 0.1],
            "boundaries": u_bounds([0.1, 0.1, 0.1]),
        },
        "temperature": {"value": 300.0, "boundaries": t_bounds(300.0)},
        "pressure": {"value": 1000.0, "boundaries": p_bounds(1000.0)},
        "catalyst_amount": 1.5e-5,
    }
    with pytest.raises(ValueError, match="Invalid composition") as err:
        COCatalyticFOAMModel(**config_error)


def test_wrong_co_from_pkl_cache():
    from osp.models.catalytic.co_catalyticfoam import COCatalyticFOAMModel

    from .utils import chem_bounds, p_bounds, t_bounds, u_bounds

    invalid_uuid = str(uuid4())
    config_error = {
        "chemical_species": [
            {"composition": "CO", "mass_fraction": 1, "boundaries": chem_bounds(1)}
        ],
        "species_from_upload": invalid_uuid,
        "velocity": {
            "value": [0.1, 0.1, 0.1],
            "boundaries": u_bounds([0.1, 0.1, 0.1]),
        },
        "temperature": {"value": 300.0, "boundaries": t_bounds(300.0)},
        "pressure": {"value": 1000.0, "boundaries": p_bounds(1000.0)},
        "catalyst_amount": 1.5e-5,
    }
    with pytest.raises(
        ValueError, match=f"Key `{invalid_uuid}` does not exist in cache."
    ) as err:
        COCatalyticFOAMModel(**config_error)


def test_mass_fraction_co():
    from osp.models.catalytic.co_catalyticfoam import COCatalyticFOAMModel

    from .utils import chem_bounds, p_bounds, t_bounds, u_bounds

    config = {
        "chemical_species": [
            {"composition": "CO", "mass_fraction": 0.5, "boundaries": chem_bounds(0.5)},
            {"composition": "O2", "mass_fraction": 0.5, "boundaries": chem_bounds(0.5)},
        ],
        "velocity": {
            "value": [0.1, 0.1, 0.1],
            "boundaries": u_bounds([0.1, 0.1, 0.1]),
        },
        "temperature": {"value": 300.0, "boundaries": t_bounds(300.0)},
        "pressure": {"value": 1000.0, "boundaries": p_bounds(1000.0)},
        "catalyst_amount": 1.5e-5,
    }
    model = COCatalyticFOAMModel(**config)
    assert model.chemical_species[0].composition == "CO"
    assert model.chemical_species[0].mass_fraction == 0.5
    assert model.chemical_species[1].composition == "O2"
    assert model.chemical_species[1].mass_fraction == 0.5
    assert model.velocity.value[0] == 0.1
    assert model.velocity.value[1] == 0.1
    assert model.velocity.value[2] == 0.1


def test_invalid_patch():
    from osp.models.catalytic.co_catalyticfoam import Boundary, COCatalyticFOAMModel

    from .utils import p_bounds, t_bounds, u_bounds

    invalid_bounds = [
        {"patch": "foo", "boundary_type": Boundary.wedge},
        {"patch": "wedge2", "boundary_type": Boundary.wedge},
        {"patch": "wedge3", "boundary_type": Boundary.wedge},
        {"patch": "wedge4", "boundary_type": Boundary.wedge},
        {"patch": "inlet", "boundary_type": Boundary.zeroGradient},
        {"patch": "outlet", "boundary_type": Boundary.fixedValue, "value": 1},
        {"patch": "inertWall", "boundary_type": Boundary.zeroGradient},
    ]
    config_error = {
        "chemical_species": [
            {"composition": "CO", "mass_fraction": 1, "boundaries": invalid_bounds},
        ],
        "velocity": {
            "value": [0.1, 0.1, 0.1],
            "boundaries": u_bounds([0.1, 0.1, 0.1]),
        },
        "temperature": {"value": 300.0, "boundaries": t_bounds(300.0)},
        "pressure": {"value": 1000.0, "boundaries": p_bounds(1000.0)},
        "catalyst_amount": 1.5e-5,
    }
    with pytest.raises(ValueError, match="Invalid patch") as err:
        COCatalyticFOAMModel(**config_error)


def test_invalid_patch_list_match():
    from osp.models.catalytic.co_catalyticfoam import Boundary, COCatalyticFOAMModel

    from .utils import p_bounds, t_bounds, u_bounds

    invalid_bounds = [
        {"patch": "wedge1", "boundary_type": Boundary.wedge},
        {"patch": "wedge2", "boundary_type": Boundary.wedge},
        {"patch": "wedge3", "boundary_type": Boundary.wedge},
        {"patch": "wedge4", "boundary_type": Boundary.wedge},
        {"patch": "inlet", "boundary_type": Boundary.zeroGradient},
        {"patch": "outlet", "boundary_type": Boundary.fixedValue, "value": 1},
        {"patch": "inertWall", "boundary_type": Boundary.zeroGradient},
    ]
    config_error = {
        "chemical_species": [
            {"composition": "CO", "mass_fraction": 1, "boundaries": invalid_bounds},
        ],
        "velocity": {
            "value": [0.1, 0.1, 0.1],
            "boundaries": u_bounds([0.1, 0.1, 0.1]),
        },
        "temperature": {"value": 300.0, "boundaries": t_bounds(300.0)},
        "pressure": {"value": 1000.0, "boundaries": p_bounds(1000.0)},
        "catalyst_amount": 1.5e-5,
    }
    with pytest.raises(ValueError, match="The provided patches for the species") as err:
        COCatalyticFOAMModel(**config_error)


def test_wrong_mass_co_upper():
    from pydantic.error_wrappers import ValidationError

    from osp.models.catalytic.co_catalyticfoam import COCatalyticFOAMModel

    from .utils import chem_bounds, p_bounds, t_bounds, u_bounds

    config_error = {
        "chemical_species": [
            {"composition": "CO", "mass_fraction": 1.3, "boundaries": chem_bounds(1.3)},
        ],
        "velocity": {
            "value": [0.1, 0.1, 0.1],
            "boundaries": u_bounds([0.1, 0.1, 0.1]),
        },
        "temperature": {"value": 300.0, "boundaries": t_bounds(300.0)},
        "pressure": {"value": 1000.0, "boundaries": p_bounds(1000.0)},
        "catalyst_amount": 1.5e-5,
    }
    with pytest.raises(
        ValidationError, match="ensure this value is less than or equal to 1"
    ) as err:
        COCatalyticFOAMModel(**config_error)


def test_wrong_mass_co_lower():
    from pydantic.error_wrappers import ValidationError

    from osp.models.catalytic.co_catalyticfoam import COCatalyticFOAMModel

    from .utils import chem_bounds, p_bounds, t_bounds, u_bounds

    config_error = {
        "chemical_species": [
            {
                "composition": "CO",
                "mass_fraction": -0.1,
                "boundaries": chem_bounds(-0.1),
            },
        ],
        "velocity": {
            "value": [0.1, 0.1, 0.1],
            "boundaries": u_bounds([0.1, 0.1, 0.1]),
        },
        "temperature": {"value": 300.0, "boundaries": t_bounds(300.0)},
        "pressure": {"value": 1000.0, "boundaries": p_bounds(1000.0)},
        "catalyst_amount": 1.5e-5,
    }
    with pytest.raises(ValidationError, match="ensure this value ") as err:
        COCatalyticFOAMModel(**config_error)


def test_wrong_mass_sum():
    from osp.models.catalytic.co_catalyticfoam import COCatalyticFOAMModel

    from .utils import chem_bounds, p_bounds, t_bounds, u_bounds

    config_error = {
        "chemical_species": [
            {
                "composition": "CO",
                "mass_fraction": 0.1,
                "boundaries": chem_bounds(0.1),
            },
            {
                "composition": "O2",
                "mass_fraction": 0.1,
                "boundaries": chem_bounds(0.1),
            },
            {
                "composition": "CO2",
                "mass_fraction": 0.3333,
                "boundaries": chem_bounds(0.3333),
            },
        ],
        "velocity": {
            "value": [0.1, 0.1, 0.1],
            "boundaries": u_bounds([0.1, 0.1, 0.1]),
        },
        "temperature": {"value": 300.0, "boundaries": t_bounds(300.0)},
        "pressure": {"value": 1000.0, "boundaries": p_bounds(1000.0)},
        "catalyst_amount": 1.5e-5,
    }
    with pytest.raises(
        ArithmeticError,
        match="The sum of `composition` of `chemical_species` must be `1.0`",
    ) as err:
        COCatalyticFOAMModel(**config_error)


def test_mass_sum():
    from osp.models.catalytic.co_catalyticfoam import COCatalyticFOAMModel

    from .utils import chem_bounds, p_bounds, t_bounds, u_bounds

    config = {
        "chemical_species": [
            {
                "composition": "CO",
                "mass_fraction": 0.3333,
                "boundaries": chem_bounds(0.3333),
            },
            {
                "composition": "O2",
                "mass_fraction": 0.3333,
                "boundaries": chem_bounds(0.3333),
            },
            {
                "composition": "CO2",
                "mass_fraction": 0.3333,
                "boundaries": chem_bounds(0.3333),
            },
        ],
        "sum_fraction_error": 4,
        "velocity": {
            "value": [0.1, 0.1, 0.1],
            "boundaries": u_bounds([0.1, 0.1, 0.1]),
        },
        "temperature": {"value": 300.0, "boundaries": t_bounds(300.0)},
        "pressure": {"value": 1000.0, "boundaries": p_bounds(1000.0)},
        "catalyst_amount": 1.5e-5,
    }
    model = COCatalyticFOAMModel(**config)
    assert model.chemical_species[0].composition == "CO"
    assert model.chemical_species[0].mass_fraction == 0.3333
    assert model.chemical_species[1].composition == "O2"
    assert model.chemical_species[1].mass_fraction == 0.3333
    assert model.chemical_species[2].composition == "CO2"
    assert model.chemical_species[2].mass_fraction == 0.3333
    assert model.velocity.value[0] == 0.1
    assert model.velocity.value[1] == 0.1
    assert model.velocity.value[2] == 0.1


def test_example_model():
    from osp.models.catalytic.co_catalyticfoam import COCatalyticFOAMModel
    from osp.models.utils.general import _get_example

    example = _get_example("catalyticFoam", "COModel.json")
    assert isinstance(example, dict)
    assert example != {}
    assert example == COCatalyticFOAMModel.Config.schema_extra["example"]
